@@ -1,5 +1,6 @@
 import distutils.version
 import functools
+import warnings
 
 import astropy
 import astropy.coordinates as coord
@@ -18,6 +19,10 @@ import pfsspy.tracing
 import pfsspy.fieldline
 # Import here to register map sources
 import pfsspy.map
+# Utility Functions
+import pfsspy.utils
+# Sample data
+import pfsspy.sample_data
 
 from .grid import Grid
 
@@ -42,13 +47,16 @@ if (distutils.version.LooseVersion(astropy.__version__) <
 _MAG_CMAP = 'RdBu'
 
 
+__all__ = ['Input', 'Output', 'load_output', 'pfss']
+
+
 class Input:
     r"""
     Input to PFSS modelling.
 
     .. warning::
         The input must be on a regularly spaced grid in :math:`\phi` and
-        :math:`s = \cos (\theta)`. See :mod:`pfsspy.coords` for more
+        :math:`s = \cos (\theta)`. See `pfsspy.grid` for more
         information on the coordinate system.
 
     Parameters
@@ -67,9 +75,17 @@ class Input:
     def __init__(self, br, nr, rss):
         if not isinstance(br, sunpy.map.GenericMap):
             raise ValueError('br must be a SunPy Map')
+        if np.any(~np.isfinite(br.data)):
+            raise ValueError('At least one value in the input is NaN or '
+                             'non-finite. The input must consist solely of '
+                             'finite values.')
+        if np.mean(br.data) > 1e-10:
+            warnings.warn('Input data has a non-zero mean. '
+                          'pfsspy will ignore this non-zero monopole term '
+                          'when calculating the PFSS solution.')
 
-        self._validate_cea(br.meta)
-        self._validate_cdelts(br)
+        pfsspy.utils.is_cea_map(br, error=True)
+        pfsspy.utils.is_full_sun_synoptic_map(br, error=True)
 
         self._map_in = br
         self.dtime = br.date
@@ -85,40 +101,6 @@ class Input:
         nphi = self.br.shape[1]
         self.grid = Grid(ns, nphi, nr, rss)
 
-    @staticmethod
-    def _validate_cea(meta):
-        for i in ('1', '2'):
-            proj = meta[f'ctype{i}'][5:8]
-            if proj != 'CEA':
-                raise ValueError(f'Projection type in CTYPE{i} keyword '
-                                 f'must be CEA (got "{proj}")')
-
-    @staticmethod
-    def _validate_cdelts(br):
-        """
-        Check that the synoptic map covers the whole Sun.
-        """
-        shape = br.data.shape
-
-        dphi = br.meta['cdelt1']
-        phi = shape[1] * dphi
-        try:
-            np.testing.assert_almost_equal(phi, 360, decimal=1)
-        except AssertionError:
-            raise ValueError('Number of points in phi direction times '
-                             'CDELT1 must be close to 360 degrees. '
-                             f'Instead got {dphi} x {shape[0]} = {phi}')
-
-        dtheta = br.meta['cdelt2']
-        theta = shape[0] * dtheta * np.pi / 2
-        try:
-            np.testing.assert_almost_equal(theta, 180, decimal=1)
-        except AssertionError:
-            raise ValueError('Number of points in theta direction times '
-                             'CDELT2 times pi/2 must be close to '
-                             '180 degrees. '
-                             f'Instead got {dtheta} x {shape[0]} * pi/2 = {theta}')
-
     @property
     def map(self):
         """
@@ -127,48 +109,6 @@ class Input:
         return self._map_in
 
 
-def carr_cea_wcs_header(dtime, shape):
-    """
-    Create a Carrington WCS header for a Cylindrical Equal Area (CEA)
-    projection. See [1]_ for information on how this is constructed.
-
-    dtime : datetime, None
-        Datetime to associate with the map.
-    shape : tuple
-        Map shape. The first entry should be number of points in longitude, the
-        second in latitude.
-
-    References
-    ----------
-    .. [1] W. T. Thompson, "Coordinate systems for solar image data",
-       https://doi.org/10.1051/0004-6361:20054262
-    """
-    # If datetime is None, put in a dummy value here to make
-    # make_fitswcs_header happy, then strip it out at the end
-    obstime = dtime or astropy.time.Time('2000-1-1')
-
-    frame_out = coord.SkyCoord(
-        0 * u.deg, 0 * u.deg, obstime=obstime,
-        frame="heliographic_carrington", observer='sun')
-    # Construct header
-    header = sunpy.map.make_fitswcs_header(
-        shape, frame_out,
-        scale=[360 / shape[0],
-               180 / shape[1]] * u.deg / u.pix,
-        reference_pixel=[(shape[0] / 2) + 0.5, (shape[1] / 2) + 0.5] * u.pix,
-        projection_code="CEA")
-
-    # Fill in these missing values
-    header['PV1_1'] = 1
-    header['PV2_1'] = 1
-    # Fix CELT for lat axis
-    header['CDELT2'] = (180 / np.pi) * (2 / shape[1])
-    # pop out the time if it isn't supplied
-    if dtime is None:
-        header.pop('date-obs')
-    return header
-
-
 class _OutOfBoundsError(RuntimeError):
     pass
 
@@ -217,6 +157,11 @@ class Output:
         Grid that the output was caclulated on.
     input_map : sunpy.map.GenericMap
         The input map.
+
+    Notes
+    -----
+    Instances of this class are intended to be created by `pfsspy.pfss`, and
+    not by users.
     '''
     def __init__(self, alr, als, alp, grid, input_map=None):
         self._alr = alr
@@ -277,6 +222,25 @@ class Output:
     def dtime(self):
         return self.input_map.date
 
+    @property
+    def bunit(self):
+        """
+        `~astropy.units.Unit` of the input map data.
+        """
+        # Note that this can be removed once pfsspy depends on sunpy>=2.1, see
+        # https://github.com/sunpy/sunpy/pull/4451
+        unit_str = self.input_map.meta.get('bunit', None)
+        if unit_str is None:
+            return
+
+        unit = u.Unit(unit_str, format='fits', parse_strict='silent')
+        if isinstance(unit, u.UnrecognizedUnit):
+            warnings.warn(f'Could not parse unit string "{unit_str}" as a valid FITS unit.\n'
+                          'See https://fits.gsfc.nasa.gov/fits_standard.html '
+                          'for the FITS unit standards.')
+            unit = None
+        return unit
+
     @property
     def source_surface_br(self):
         """
@@ -387,9 +351,6 @@ class Output:
         import scipy.integrate
 
         direction = np.sign(dt)
-        dt = np.abs(dt)
-        t = 0.0
-        xout = np.atleast_2d(start_point.copy())
 
         def finish_integration(t, coord):
             r = np.linalg.norm(coord)
@@ -413,7 +374,7 @@ class Output:
         return xout
 
     @property
-    def al(self):
+    def _al(self):
         """
         Vector potential times cell edge lenghts.
 
@@ -425,6 +386,12 @@ class Output:
     def bc(self):
         """
         B on the centres of the cell faces.
+
+        Returns
+        -------
+        br
+        btheta
+        bphi
         """
         br, bs, bp, Sbr, Sbs, Sbp = self._common_b()
         # Remove area factors:
@@ -497,12 +464,14 @@ class Output:
         rg = self.grid.rg
         sg = self.grid.sg
 
-        alr, als, alp = self.al
+        alr, als, alp = self._al
 
+        # Centre of cells in rho (including ghost cells)
         rc = np.linspace(-0.5 * dr, np.log(rss) + 0.5 * dr, nr + 2)
         rrc = np.exp(rc)
         thc = np.zeros(ns + 2) - 1
         thc[1:-1] = np.arccos(sc)
+        # Centre of cells in phi (including ghost cells)
         pc = np.linspace(-0.5 * dp, 2 * np.pi + 0.5 * dp, nphi + 2)
 
         # Required face normals:
@@ -532,40 +501,40 @@ class Output:
         Sbs = np.zeros((ns + 1, nr + 2))
         for k in range(nr + 2):
             for j in range(1, ns):
-                Sbs[j,k] = 0.5*np.exp(2*rc[k] - dr)*dp*(np.exp(2*dr)-1)*np.sqrt(1 - sg[j]**2)
-            Sbs[0,k] = Sbs[1,k]
-            Sbs[-1,k] = Sbs[-2,k]
-        Sbp = np.zeros((ns+2,nr+2))
-        for k in range(nr+2):
-            for j in range(1,ns+1):
-                Sbp[j,k] = 0.5*np.exp(2*rc[k] - dr)*(np.exp(2*dr) - 1)*(np.arcsin(sg[j]) - np.arcsin(sg[j-1]))
-            Sbp[0,k] = Sbp[1,k]
-            Sbp[-1,k] = Sbp[-2,k]
+                Sbs[j, k] = 0.5 * np.exp(2 * rc[k] - dr) * dp * (np.exp(2 * dr) - 1) * np.sqrt(1 - sg[j]**2)
+            Sbs[0, k] = Sbs[1, k]
+            Sbs[-1, k] = Sbs[-2, k]
+        Sbp = np.zeros((ns + 2, nr + 2))
+        for k in range(nr + 2):
+            for j in range(1, ns + 1):
+                Sbp[j, k] = 0.5 * np.exp(2 * rc[k] - dr) * (np.exp(2 * dr) - 1) * (np.arcsin(sg[j]) - np.arcsin(sg[j - 1]))
+            Sbp[0, k] = Sbp[1, k]
+            Sbp[-1, k] = Sbp[-2, k]
 
         # Compute br*Sbr, bs*Sbs, bp*Sbp at cell centres by Stokes theorem:
-        br = np.zeros((nphi+2,ns+2,nr+1))
-        bs = np.zeros((nphi+2,ns+1,nr+2))
-        bp = np.zeros((nphi+1,ns+2,nr+2))
-        br[1:-1,1:-1,:] = als[1:,:,:] - als[:-1,:,:] + alp[:,:-1,:] - alp[:,1:,:]
-        bs[1:-1,:,1:-1] = alp[:,:,1:] - alp[:,:,:-1]
-        bp[:,1:-1,1:-1] = als[:,:,:-1] - als[:,:,1:]
+        br = np.zeros((nphi + 2, ns + 2, nr + 1))
+        bs = np.zeros((nphi + 2, ns + 1, nr + 2))
+        bp = np.zeros((nphi + 1, ns + 2, nr + 2))
+        br[1:-1, 1:-1, :] = als[1:, :, :] - als[:-1, :, :] + alp[:, :-1, :] - alp[:, 1:, :]
+        bs[1:-1, :, 1:-1] = alp[:, :, 1:] - alp[:, :, :-1]
+        bp[:, 1:-1, 1:-1] = als[:, :, :-1] - als[:, :, 1:]
 
         # Fill ghost values with boundary conditions:
         # - zero-gradient at outer boundary:
-        bs[1:-1,:,-1] = 2*bs[1:-1,:,-2] - bs[1:-1,:,-3]
-        bp[:,1:-1,-1] = 2*bp[:,1:-1,-2] - bp[:,1:-1,-3]
+        bs[1:-1, :, -1] = 2 * bs[1:-1, :, -2] - bs[1:-1, :, -3]
+        bp[:, 1:-1, -1] = 2 * bp[:, 1:-1, -2] - bp[:, 1:-1, -3]
         # - periodic in phi:
-        bs[0,:,:] = bs[-2,:,:]
-        bs[-1,:,:] = bs[1,:,:]
-        br[0,:,:] = br[-2,:,:]
-        br[-1,:,:] = br[1,:,:]
+        bs[0, :, :] = bs[-2, :, :]
+        bs[-1, :, :] = bs[1, :, :]
+        br[0, :, :] = br[-2, :, :]
+        br[-1, :, :] = br[1, :, :]
         # js = jp = 0 at photosphere:
-        for i in range(nphi+1):
+        for i in range(nphi + 1):
             bp[i,:,0] = Sbp[:,0]/dnp[:,0]*(bp[i,:,1]*dnp[:,1]/Sbp[:,1] + br[i,:,0]*dnr[:]/Sbr[:,0] - br[i+1,:,0]*dnr[:]/Sbr[:,0])
-        for i in range(nphi+2):
+        for i in range(nphi + 2):
             bs[i,:,0] = Sbs[:,0]/dns[:,0]*(bs[i,:,1]*dns[:,1]/Sbs[:,1] + br[i,:-1,0]*dnr[:-1]/Sbr[:-1,0] - br[i,1:,0]*dnr[1:]/Sbr[1:,0])
         # - polar boundaries as in dumfric:
-        for i in range(nphi+2):
+        for i in range(nphi + 2):
             i1 = (i + nphi//2) % nphi
             br[i,-1,:] = br[i1,-2,:]
             br[i,0,:] = br[i1,1,:]
@@ -579,6 +548,104 @@ class Output:
         self._common_b_cache = br, bs, bp, Sbr, Sbs, Sbp
         return self._common_b_cache
 
+    def get_bvec(self, coords, out_type="spherical"):
+        """
+        Evaluate magnetic vectors in pfss model.
+
+        Method which takes an arbitrary astropy SkyCoord and
+        returns a numpy array containing magnetic field vectors
+        evaluated from the parent pfsspy.Output pfss model at
+        the locations specified by the SkyCoords
+
+        Parameters
+        ----------
+        coords : `astropy.SkyCoord`
+            An arbitary point or set of points (length N >= 1)
+            in the PFSS model domain (1Rs < r < Rss)
+
+        out_type : str, optional
+            Takes values 'spherical' (default) or 'cartesian'
+            and specifies whether the output vector is in
+            spherical coordinates (B_r,B_theta,B_phi) or
+            cartesian (B_x,B_y,B_z)
+
+        Returns
+        -------
+        bvec : ndarray
+            Magnetic field vectors at the requested locations
+            ndarray.shape = (N,3), units nT)
+
+        Notes
+        -----
+        The output coordinate system is defined by the input
+        magnetogram with x-z plane equivalent to the plane
+        containing the Carrington meridian (0 deg longitude)
+
+        The spherical coordinates follow the physics convention:
+        https://upload.wikimedia.org/wikipedia/commons/thumb/4/4f/3D_Spherical.svg/240px-3D_Spherical.svg.png)
+        Therefore the polar angle (theta) is the co-latitude, rather
+        than the latitude, with range 0 (north pole) to 180 degrees
+        (south pole)
+
+        The conversion which relates the spherical and cartesian
+        coordinates is as follows:
+
+        .. math:: B_R = sin\\theta cos\\phi B_x + sin\\theta sin\\phi B_y + cos\\theta B_z
+        .. math:: B_\\theta = cos\\theta cos\\phi B_x + cos\\theta sin\\phi B_y - sin\\theta B_z
+        .. math:: B_\\phi = -sin\\phi B_x + cos\\phi B_y
+
+        The above equations may be written as a (3x3) matrix and
+        inverted to retrieve the inverse transformation (cartesian from spherical)
+        """
+
+        # Assert skycoord is type astropy.coordinates.SkyCoord
+        if not isinstance(coords, coord.SkyCoord):
+            raise ValueError("coords must be of type "
+                             "astropy.coordinates.SkyCoord")
+
+        # Ensure representation type is spherical for input to interpolator
+        coords.representation_type = "spherical"
+
+        # Check coord_type is cartesian or spherical
+        if out_type not in ["cartesian", "spherical"]:
+            raise ValueError("out_type must be 'cartesian' or 'spherical' "
+                             f"(got {out_type})")
+
+        # Raise warning if input skycoord obstime does not match
+        # self.coordinate_frame.obstime
+        if np.any(self.coordinate_frame.obstime.to_datetime() !=
+                  coords.obstime.to_datetime()):
+            warnings.warn("The obstime of one of more input coordinates "
+                          "do not match the pfss model obstime.")
+
+        # Convert SkyCoord to pfsspy.Output coordinate frame
+        coords.transform_to(self.coordinate_frame)
+
+        # Do interpolation (returns cartesian vector)
+        bvecs = self._brgi(np.array([coords.lon.to("rad").value,
+                                     np.sin(coords.lat).value,
+                                     np.log(coords.radius.to("R_sun").value)]).T
+                           )
+
+        # Convert to spherical if requested
+        if out_type == "spherical":
+            # Generate vector of 3x3 rotation matrices
+            M = np.array([
+                [np.cos(coords.lat).value * np.cos(coords.lon).value,
+                 np.cos(coords.lat).value * np.sin(coords.lon).value,
+                 np.sin(coords.lat).value],
+                [np.sin(coords.lat).value * np.cos(coords.lon).value,
+                 np.sin(coords.lat).value * np.sin(coords.lon).value,
+                 -np.cos(coords.lat).value],
+                [-np.sin(coords.lon).value,
+                 np.cos(coords.lon).value,
+                 np.zeros(len(coords))],
+            ])
+            bvecs = np.array([np.dot(M_.T,v) for M_,v in zip(M.T,bvecs)])
+        if self.bunit is not None:
+            bvecs *= self.bunit
+        return bvecs
+
 
 def _eigh(A):
     return np.linalg.eigh(A)
@@ -586,7 +653,13 @@ def _eigh(A):
 
 def _compute_r_term(m, k, ns, Q, brt, lam, ffm, nr, ffp, psi, psir):
     for l in range(ns):
-        # - sum c_{lm} + d_{lm}
+        # Ignore the l=0 and m=0 term; for a globally divergence free field
+        # this term is zero anyway, but numerically it may be small which
+        # causes numerical issues when solving for c, d
+        if l == 0 and m == 0:
+            continue
+        # - sum (c_{lm} + d_{lm}) * lam_{l}
+        # lam[l] is small so this blows up
         cdlm = np.dot(Q[:, l], brt[:, m]) / lam[l]
         # - ratio c_{lm}/d_{lm} [numerically safer this way up]
         ratio = (ffm[l]**(nr - 1) - ffm[l]**nr) / (ffp[l]**nr - ffp[l]**(nr - 1))
@@ -630,10 +703,6 @@ def pfss(input):
     (equally spaced in :math:`\rho = \ln(r/r_{sun})`,
     :math:`s= \cos(\theta)`, and :math:`p=\phi`).
 
-    The output should have zero current to machine precision,
-    when computed with the DuMFriC staggered discretization.
-
-
     Parameters
     ----------
     input : :class:`Input`
@@ -642,12 +711,20 @@ def pfss(input):
     Returns
     -------
     out : :class:`Output`
+
+    Notes
+    -----
+    In order to avoid numerical issues, the monopole term (which should be zero
+    for a physical magnetic field anyway) is explicitly excluded from the
+    solution.
+
+    The output should have zero current to machine precision,
+    when computed with the DuMFriC staggered discretization.
     """
     br0 = input.br
     nr = input.grid.nr
     ns = input.grid.ns
     nphi = input.grid.nphi
-    rss = input.grid.rss
 
     # Coordinates:
     ds = input.grid.ds
@@ -676,8 +753,8 @@ def pfss(input):
     # - create off-diagonal part of the matrix:
     A = np.zeros((ns, ns))
     for j in range(ns - 1):
-        A[j, j+1] = -Vg[j+1]
-        A[j+1, j] = A[j, j+1]
+        A[j, j + 1] = -Vg[j + 1]
+        A[j + 1, j] = A[j, j + 1]
     # - term required for m-dependent part of matrix:
     mu = np.fft.fftfreq(nphi)
     mu = 4 * np.sin(np.pi * mu)**2
@@ -712,6 +789,7 @@ def pfss(input):
     psi = np.real(np.fft.ifft(psi, axis=2))
 
     # Hence compute vector potential [note index order, for netcdf]:
+    # (note that alr is zero by definition)
     alr = np.zeros((nphi + 1, ns + 1, nr))
     als = np.zeros((nphi + 1, ns, nr + 1))
     alp = np.zeros((nphi, ns + 1, nr + 1))
